package jdk.internal.org.objectweb.asm.util;

import jdk.internal.org.objectweb.asm.Attribute;
import jdk.internal.org.objectweb.asm.TypePath;
import jdk.internal.org.objectweb.asm.AnnotationVisitor;
import jdk.internal.org.objectweb.asm.MethodVisitor;
import jdk.internal.org.objectweb.asm.FieldVisitor;
import java.util.HashMap;
import jdk.internal.org.objectweb.asm.tree.analysis.Frame;
import jdk.internal.org.objectweb.asm.tree.TryCatchBlockNode;
import jdk.internal.org.objectweb.asm.tree.analysis.BasicValue;
import java.util.Iterator;
import java.util.List;
import jdk.internal.org.objectweb.asm.tree.analysis.Analyzer;
import jdk.internal.org.objectweb.asm.tree.analysis.SimpleVerifier;
import jdk.internal.org.objectweb.asm.tree.MethodNode;
import java.util.ArrayList;
import jdk.internal.org.objectweb.asm.Type;
import jdk.internal.org.objectweb.asm.tree.ClassNode;
import java.io.PrintWriter;
import java.io.FileInputStream;
import jdk.internal.org.objectweb.asm.ClassReader;
import jdk.internal.org.objectweb.asm.Label;
import java.util.Map;
import jdk.internal.org.objectweb.asm.ClassVisitor;

public class CheckClassAdapter extends ClassVisitor
{
    private int version;
    private boolean start;
    private boolean source;
    private boolean outer;
    private boolean end;
    private Map<Label, Integer> labels;
    private boolean checkDataFlow;
    
    public static void main(String[] array) throws Exception {
        if (array.length != 1) {
            System.err.println("Verifies the given class.");
            System.err.println("Usage: CheckClassAdapter <fully qualified class name or class file name>");
            return;
        }
        ClassReader classReader;
        if (array[0].endsWith(".class")) {
            classReader = new ClassReader(new FileInputStream(array[0]));
        }
        else {
            classReader = new ClassReader(array[0]);
        }
        verify(classReader, false, new PrintWriter(System.err));
        return;
    }
    
    public static void verify(ClassReader classReader, ClassLoader classLoader, boolean b, PrintWriter s) {
        ClassNode classNode = new ClassNode();
        classReader.accept(new CheckClassAdapter(classNode, false), 2);
        Type currentSuperClass = (classNode.superName == null) ? null : Type.getObjectType(classNode.superName);
        List<Object> methods = classNode.methods;
        ArrayList<Type> currentClassInterfaces = new ArrayList();
        Iterator<String> iterator = classNode.interfaces.iterator();
        while (iterator.hasNext()) {
            currentClassInterfaces.add(Type.getObjectType((String)iterator.next()));
        }
        for (int i = 0; i < methods.size(); ++i) {
            MethodNode method = (MethodNode)methods.get(i);
            SimpleVerifier interpreter = new SimpleVerifier(Type.getObjectType(classNode.name), currentSuperClass, currentClassInterfaces, (classNode.access & 0x200) != 0x0);
            Analyzer o = new Analyzer(interpreter);
            if (classLoader != null) {
                interpreter.setClassLoader(classLoader);
            }
            try {
                o.analyze(classNode.name, method);
                if (!b) {
                    continue;
                }
            }
            catch (Exception ex) {
                ex.printStackTrace(s);
            }
            printAnalyzerResult(method, o, s);
        }
        s.flush();
    }
    
    public static void verify(ClassReader classReader, boolean b, PrintWriter printWriter) {
        verify(classReader, null, b, printWriter);
    }
    
    static void printAnalyzerResult(MethodNode methodNode, Analyzer<BasicValue> analyzer, PrintWriter printWriter) {
        Frame[] frames = analyzer.getFrames();
        Textifier textifier = new Textifier();
        TraceMethodVisitor methodVisitor = new TraceMethodVisitor(textifier);
        printWriter.println(methodNode.name + methodNode.desc);
        for (int i = 0; i < methodNode.instructions.size(); ++i) {
            methodNode.instructions.get(i).accept(methodVisitor);
            StringBuilder obj = new StringBuilder();
            Frame<Object> frame = frames[i];
            if (frame == null) {
                obj.append('?');
            }
            else {
                for (int j = 0; j < frame.getLocals(); ++j) {
                    obj.append(getShortName(((BasicValue)frame.getLocal(j)).toString())).append(' ');
                }
                obj.append(" : ");
                for (int k = 0; k < frame.getStackSize(); ++k) {
                    obj.append(getShortName(((BasicValue)frame.getStack(k)).toString())).append(' ');
                }
            }
            while (obj.length() < methodNode.maxStack + methodNode.maxLocals + 1) {
                obj.append(' ');
            }
            printWriter.print(Integer.toString(i + 100000).substring(1));
            printWriter.print(" " + obj + " : " + textifier.text.get(textifier.text.size() - 1));
        }
        for (int l = 0; l < methodNode.tryCatchBlocks.size(); ++l) {
            ((TryCatchBlockNode)methodNode.tryCatchBlocks.get(l)).accept(methodVisitor);
            printWriter.print(" " + textifier.text.get(textifier.text.size() - 1));
        }
        printWriter.println();
    }
    
    private static String getShortName(String s) {
        int lastIndex = s.lastIndexOf(47);
        int length = s.length();
        if (s.charAt(length - 1) == ';') {
            --length;
        }
        return (lastIndex == -1) ? s : s.substring(lastIndex + 1, length);
    }
    
    public CheckClassAdapter(ClassVisitor classVisitor) {
        this(classVisitor, true);
    }
    
    public CheckClassAdapter(ClassVisitor classVisitor, boolean b) {
        this(327680, classVisitor, b);
        if (this.getClass() != CheckClassAdapter.class) {
            throw new IllegalStateException();
        }
        return;
    }
    
    protected CheckClassAdapter(int api, ClassVisitor classVisitor, boolean checkDataFlow) {
        super(api, classVisitor);
        this.labels = new HashMap();
        this.checkDataFlow = checkDataFlow;
    }
    
    @Override
    public void visit(int version, int access, String s, String signature, String s2, String[] interfaces) {
        if (this.start) {
            throw new IllegalStateException("visit must be called only once");
        }
        this.start = true;
        this.checkState();
        checkAccess(access, 423473);
        if (s == null || !s.endsWith("package-info")) {
            CheckMethodAdapter.checkInternalName(s, "class name");
        }
        if ("java/lang/Object".equals(s)) {
            if (s2 != null) {
                throw new IllegalArgumentException("The super class name of the Object class must be 'null'");
            }
        }
        else {
            CheckMethodAdapter.checkInternalName(s2, "super class name");
        }
        if (signature != null) {
            checkClassSignature(signature);
        }
        if ((access & 0x200) != 0x0 && !"java/lang/Object".equals(s2)) {
            throw new IllegalArgumentException("The super class name of interfaces must be 'java/lang/Object'");
        }
        if (interfaces != null) {
            for (int i = 0; i < interfaces.length; ++i) {
                CheckMethodAdapter.checkInternalName(interfaces[i], "interface name at index " + i);
            }
        }
        super.visit(this.version = version, access, s, signature, s2, interfaces);
        return;
    }
    
    @Override
    public void visitSource(String source, String debug) {
        this.checkState();
        if (this.source) {
            throw new IllegalStateException("visitSource can be called only once.");
        }
        this.source = true;
        super.visitSource(source, debug);
        return;
    }
    
    @Override
    public void visitOuterClass(String owner, String name, String descriptor) {
        this.checkState();
        if (this.outer) {
            throw new IllegalStateException("visitOuterClass can be called only once.");
        }
        this.outer = true;
        if (owner == null) {
            throw new IllegalArgumentException("Illegal outer class owner");
        }
        if (descriptor != null) {
            CheckMethodAdapter.checkMethodDesc(descriptor);
        }
        super.visitOuterClass(owner, name, descriptor);
        return;
    }
    
    @Override
    public void visitInnerClass(String name, String outerName, String innerName, int access) {
        this.checkState();
        CheckMethodAdapter.checkInternalName(name, "class name");
        if (outerName != null) {
            CheckMethodAdapter.checkInternalName(outerName, "outer class name");
        }
        if (innerName != null) {
            int index;
            for (index = 0; index < innerName.length() && Character.isDigit(innerName.charAt(index)); ++index) {}
            if (index == 0 || index < innerName.length()) {
                CheckMethodAdapter.checkIdentifier(innerName, index, -1, "inner class name");
            }
        }
        checkAccess(access, 30239);
        super.visitInnerClass(name, outerName, innerName, access);
    }
    
    @Override
    public FieldVisitor visitField(int access, String name, String descriptor, String signature, Object value) {
        this.checkState();
        checkAccess(access, 413919);
        CheckMethodAdapter.checkUnqualifiedName(this.version, name, "field name");
        CheckMethodAdapter.checkDesc(descriptor, false);
        if (signature != null) {
            checkFieldSignature(signature);
        }
        if (value != null) {
            CheckMethodAdapter.checkConstant(value);
        }
        return new CheckFieldAdapter(super.visitField(access, name, descriptor, signature, value));
    }
    
    @Override
    public MethodVisitor visitMethod(int n, String s, String s2, String s3, String[] array) {
        this.checkState();
        checkAccess(n, 400895);
        if (!"<init>".equals(s) && !"<clinit>".equals(s)) {
            CheckMethodAdapter.checkMethodIdentifier(this.version, s, "method name");
        }
        CheckMethodAdapter.checkMethodDesc(s2);
        if (s3 != null) {
            checkMethodSignature(s3);
        }
        if (array != null) {
            for (int i = 0; i < array.length; ++i) {
                CheckMethodAdapter.checkInternalName(array[i], "exception name at index " + i);
            }
        }
        Object o;
        if (this.checkDataFlow) {
            o = new CheckMethodAdapter(n, s, s2, super.visitMethod(n, s, s2, s3, array), this.labels);
        }
        else {
            o = new CheckMethodAdapter(super.visitMethod(n, s, s2, s3, array), this.labels);
        }
        o.version = this.version;
        return o;
    }
    
    @Override
    public AnnotationVisitor visitAnnotation(String descriptor, boolean visible) {
        this.checkState();
        CheckMethodAdapter.checkDesc(descriptor, false);
        return new CheckAnnotationAdapter(super.visitAnnotation(descriptor, visible));
    }
    
    @Override
    public AnnotationVisitor visitTypeAnnotation(int typeRef, TypePath typePath, String descriptor, boolean visible) {
        this.checkState();
        int i = typeRef >>> 24;
        if (i != 0 && i != 17 && i != 16) {
            throw new IllegalArgumentException("Invalid type reference sort 0x" + Integer.toHexString(i));
        }
        checkTypeRefAndPath(typeRef, typePath);
        CheckMethodAdapter.checkDesc(descriptor, false);
        return new CheckAnnotationAdapter(super.visitTypeAnnotation(typeRef, typePath, descriptor, visible));
    }
    
    @Override
    public void visitAttribute(Attribute attribute) {
        this.checkState();
        if (attribute == null) {
            throw new IllegalArgumentException("Invalid attribute (must not be null)");
        }
        super.visitAttribute(attribute);
        return;
    }
    
    @Override
    public void visitEnd() {
        this.checkState();
        this.end = true;
        super.visitEnd();
    }
    
    private void checkState() {
        if (!this.start) {
            throw new IllegalStateException("Cannot visit member before visit has been called.");
        }
        if (this.end) {
            throw new IllegalStateException("Cannot visit member after visitEnd has been called.");
        }
        return;
    }
    
    static void checkAccess(int i, int n) {
        if ((i & ~n) != 0x0) {
            throw new IllegalArgumentException("Invalid access flags: " + i);
        }
        if (((((i & 0x1) != 0x0) + ((i & 0x2) != 0x0) + ((i & 0x4) != 0x0)) ? 1 : 0) > 1) {
            throw new IllegalArgumentException("public private and protected are mutually exclusive: " + i);
        }
        if (((((i & 0x10) != 0x0) + ((i & 0x400) != 0x0)) ? 1 : 0) > 1) {
            throw new IllegalArgumentException("final and abstract are mutually exclusive: " + i);
        }
        return;
    }
    
    public static void checkClassSignature(String str) {
        int checkFormalTypeParameters = 0;
        if (getChar(str, 0) == '<') {
            checkFormalTypeParameters = checkFormalTypeParameters(str, checkFormalTypeParameters);
        }
        int i;
        for (i = checkClassTypeSignature(str, checkFormalTypeParameters); getChar(str, i) == 'L'; i = checkClassTypeSignature(str, i)) {}
        if (i != str.length()) {
            throw new IllegalArgumentException(str + ": error at index " + i);
        }
        return;
    }
    
    public static void checkMethodSignature(String str) {
        int checkFormalTypeParameters = 0;
        if (getChar(str, 0) == '<') {
            checkFormalTypeParameters = checkFormalTypeParameters(str, checkFormalTypeParameters);
        }
        int n;
        for (n = checkChar('(', str, checkFormalTypeParameters); "ZCBSIFJDL[T".indexOf((int)getChar(str, n)) != -1; n = checkTypeSignature(str, n)) {}
        int i = checkChar(')', str, n);
        if (getChar(str, i) == 'V') {
            ++i;
        }
        else {
            i = checkTypeSignature(str, i);
        }
        while (getChar(str, i) == '^') {
            ++i;
            if (getChar(str, i) == 'L') {
                i = checkClassTypeSignature(str, i);
            }
            else {
                i = checkTypeVariableSignature(str, i);
            }
        }
        if (i != str.length()) {
            throw new IllegalArgumentException(str + ": error at index " + i);
        }
        return;
    }
    
    public static void checkFieldSignature(String str) {
        int checkFieldTypeSignature = checkFieldTypeSignature(str, 0);
        if (checkFieldTypeSignature != str.length()) {
            throw new IllegalArgumentException(str + ": error at index " + checkFieldTypeSignature);
        }
        return;
    }
    
    static void checkTypeRefAndPath(int i, TypePath typePath) {
        int n;
        switch (i >>> 24) {
            case 0:
            case 1:
            case 22: {
                n = -65536;
                break;
            }
            case 19:
            case 20:
            case 21:
            case 64:
            case 65:
            case 67:
            case 68:
            case 69:
            case 70: {
                n = -16777216;
                break;
            }
            case 16:
            case 17:
            case 18:
            case 23:
            case 66: {
                n = -256;
                break;
            }
            case 71:
            case 72:
            case 73:
            case 74:
            case 75: {
                n = -16776961;
                break;
            }
            default: {
                throw new IllegalArgumentException("Invalid type reference sort 0x" + Integer.toHexString(i >>> 24));
            }
        }
        if ((i & ~n) != 0x0) {
            throw new IllegalArgumentException("Invalid type reference 0x" + Integer.toHexString(i));
        }
        if (typePath != null) {
            int j = 0;
            while (j < typePath.getLength()) {
                int step = typePath.getStep(j);
                if (step != 0 && step != 1 && step != 3 && step != 2) {
                    throw new IllegalArgumentException("Invalid type path step " + j + " in " + typePath);
                }
                if (step != 3 && typePath.getStepArgument(j) != 0) {
                    throw new IllegalArgumentException("Invalid type path step argument for step " + j + " in " + typePath);
                }
                ++j;
            }
        }
        return;
    }
    
    private static int checkFormalTypeParameters(String s, int n) {
        for (n = checkChar('<', s, n), n = checkFormalTypeParameter(s, n); getChar(s, n) != '>'; n = checkFormalTypeParameter(s, n)) {}
        return n + 1;
    }
    
    private static int checkFormalTypeParameter(String s, int n) {
        n = checkIdentifier(s, n);
        n = checkChar(':', s, n);
        if ("L[T".indexOf((int)getChar(s, n)) != -1) {
            n = checkFieldTypeSignature(s, n);
        }
        while (getChar(s, n) == ':') {
            n = checkFieldTypeSignature(s, n + 1);
        }
        return n;
    }
    
    private static int checkFieldTypeSignature(String s, int n) {
        switch (getChar(s, n)) {
            case 'L': {
                return checkClassTypeSignature(s, n);
            }
            case '[': {
                return checkTypeSignature(s, n + 1);
            }
            default: {
                return checkTypeVariableSignature(s, n);
            }
        }
    }
    
    private static int checkClassTypeSignature(String s, int n) {
        for (n = checkChar('L', s, n), n = checkIdentifier(s, n); getChar(s, n) == '/'; n = checkIdentifier(s, n + 1)) {}
        if (getChar(s, n) == '<') {
            n = checkTypeArguments(s, n);
        }
        while (getChar(s, n) == '.') {
            n = checkIdentifier(s, n + 1);
            if (getChar(s, n) != '<') {
                continue;
            }
            n = checkTypeArguments(s, n);
        }
        return checkChar(';', s, n);
    }
    
    private static int checkTypeArguments(String s, int n) {
        for (n = checkChar('<', s, n), n = checkTypeArgument(s, n); getChar(s, n) != '>'; n = checkTypeArgument(s, n)) {}
        return n + 1;
    }
    
    private static int checkTypeArgument(String s, int n) {
        int char1 = getChar(s, n);
        if (char1 == '*') {
            return n + 1;
        }
        if (char1 == '+' || char1 == '-') {
            ++n;
        }
        return checkFieldTypeSignature(s, n);
    }
    
    private static int checkTypeVariableSignature(String s, int n) {
        n = checkChar('T', s, n);
        n = checkIdentifier(s, n);
        return checkChar(';', s, n);
    }
    
    private static int checkTypeSignature(String s, int n) {
        switch (getChar(s, n)) {
            case 'B':
            case 'C':
            case 'D':
            case 'F':
            case 'I':
            case 'J':
            case 'S':
            case 'Z': {
                return n + 1;
            }
            default: {
                return checkFieldTypeSignature(s, n);
            }
        }
    }
    
    private static int checkIdentifier(String str, int i) {
        if (!Character.isJavaIdentifierStart(getChar(str, i))) {
            throw new IllegalArgumentException(str + ": identifier expected at index " + i);
        }
        ++i;
        while (Character.isJavaIdentifierPart(getChar(str, i))) {
            ++i;
        }
        return i;
    }
    
    private static int checkChar(char c, String str, int i) {
        if (getChar(str, i) == c) {
            return i + 1;
        }
        throw new IllegalArgumentException(str + ": '" + c + "' expected at index " + i);
    }
    
    private static char getChar(String s, int index) {
        return (index < s.length()) ? s.charAt(index) : '\0';
    }
}
